;; # emacs -q -L /home/albinus/src/tramp-2-7-stable/lisp -l ~/src/tramp-2-7-stable/test-74105.el '/ssh:|sudo::'

(require 'tramp)
(require 'auth-source)
(require 'password-cache)
;(setq auth-sources nil)
(setq debug-on-error t tramp-verbose 10)
(add-to-list 'tramp-connection-properties '(nil "session-timeout" 8))
(dolist (fn (append (mapcar #'intern (all-completions "password-cache-" obarray #'functionp))
		    (mapcar #'intern (all-completions "password-read" obarray #'functionp))
		    (mapcar #'intern (all-completions "auth-source-" obarray #'functionp))
		    '(tramp-read-passwd tramp-clear-passwd)))
  (trace-function-background fn))
